#include "stdafx.h"
#include "clf_input.h"

static HRESULT CreateJoystickDevice( HWND hDlg );
static HRESULT CreateKeyboardDevice( HWND hDlg );
static HRESULT CreateMouseDevice( HWND hDlg );

static BOOL CALLBACK    EnumObjectsCallback( const DIDEVICEOBJECTINSTANCE* pdidoi, VOID* pContext );
static BOOL CALLBACK    EnumJoysticksCallback( const DIDEVICEINSTANCE* pdidInstance, VOID* pContext );

static HRESULT UpdateJoystick( HWND hDlg );
static HRESULT ReadImmediateKeyboard( HWND hDlg );
static HRESULT ReadBufferedKeyboard( HWND hDlg );
static HRESULT ReadImmediateMouse( HWND hDlg );
static HRESULT ReadBufferedMouse( HWND hDlg );

struct DI_ENUM_CONTEXT
{
	DIJOYCONFIG* pPreferredJoyCfg;
	bool bPreferredJoyCfgValid;
};

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------
#define SAFE_DELETE(p)  { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_RELEASE(p) { if(p) { (p)->Release(); (p)=NULL; } }
#define SAMPLE_BUFFER_SIZE 8  // arbitrary number of buffer elements
#define MOUSE_SAMPLE_BUFFER_SIZE 16

static LPDIRECTINPUT8       g_pDI       = NULL;         
static LPDIRECTINPUTDEVICE8 g_pJoystick = NULL;     
static LPDIRECTINPUTDEVICE8 g_pKeyboard = NULL;
static LPDIRECTINPUTDEVICE8 g_pMouse    = NULL;

static DIJOYSTATE2 js;           // DInput joystick state 

static DIMOUSESTATE2 dims2;      // DirectInput mouse state structure immediate
static DIDEVICEOBJECTDATA didod_mouse[ MOUSE_SAMPLE_BUFFER_SIZE ];  // Receives buffered data 
static DWORD              dwElementsMouse;

static BYTE    diks[256];   // DirectInput keyboard state buffer immediate
static DIDEVICEOBJECTDATA didod_keybd[ SAMPLE_BUFFER_SIZE ];  // Receives buffered data 
static DWORD              dwElementsKeybd;

BOOL    bExclusive         = TRUE;
BOOL    bForeground        = TRUE;
BOOL    bImmediate         = TRUE;
BOOL    bDisableWindowsKey = TRUE;

//////////////////////////////////////////////////////////////////////////
// initialize
int InitDirectInput( HWND hDlg, bool bImm, bool bKbd, bool bMouse, bool bJoy)
{
	HRESULT hr;

	// Register with the DirectInput subsystem and get a pointer
	// to a IDirectInput interface we can use.
	// Create a DInput object
	if( FAILED( hr = DirectInput8Create( GetModuleHandle(NULL), DIRECTINPUT_VERSION, 
		IID_IDirectInput8, (VOID**)&g_pDI, NULL ) ) )
		return 0;

	bImmediate = bImm;

	if (bJoy)
		CreateJoystickDevice( hDlg );
	if (bKbd)
		CreateKeyboardDevice( hDlg );
	if (bMouse)
		CreateMouseDevice( hDlg );

	return 1;
}

//////////////////////////////////////////////////////////////////////////
// terminate
void FreeDirectInput()
{
	if( g_pJoystick ) 
		g_pJoystick->Unacquire();
	if( g_pKeyboard ) 
		g_pKeyboard->Unacquire();
	if( g_pMouse ) 
		g_pMouse->Unacquire();

	SAFE_RELEASE( g_pJoystick );
	SAFE_RELEASE( g_pKeyboard );
	SAFE_RELEASE( g_pMouse );
	SAFE_RELEASE( g_pDI );
}

//////////////////////////////////////////////////////////////////////
// joystick
static HRESULT CreateJoystickDevice( HWND hDlg )
{
	HRESULT hr;

	DIJOYCONFIG PreferredJoyCfg = {0};
	DI_ENUM_CONTEXT enumContext;
	enumContext.pPreferredJoyCfg = &PreferredJoyCfg;
	enumContext.bPreferredJoyCfgValid = false;

	IDirectInputJoyConfig8* pJoyConfig = NULL;  
	if( FAILED( hr = g_pDI->QueryInterface( IID_IDirectInputJoyConfig8, (void **) &pJoyConfig ) ) )
		return hr;

	PreferredJoyCfg.dwSize = sizeof(PreferredJoyCfg);
	if( SUCCEEDED( pJoyConfig->GetConfig(0, &PreferredJoyCfg, DIJC_GUIDINSTANCE ) ) ) // This function is expected to fail if no joystick is attached
		enumContext.bPreferredJoyCfgValid = true;
	SAFE_RELEASE( pJoyConfig );

	// Look for a simple joystick we can use for this sample program.
	if( FAILED( hr = g_pDI->EnumDevices( DI8DEVCLASS_GAMECTRL, 
								EnumJoysticksCallback, &enumContext, DIEDFL_ATTACHEDONLY ) ) )
		return hr;

	// Make sure we got a joystick
	if( NULL == g_pJoystick )
	{
		return S_OK;
	}

	// Set the data format to "simple joystick" - a predefined data format 
	//
	// A data format specifies which controls on a device we are interested in,
	// and how they should be reported. This tells DInput that we will be
	// passing a DIJOYSTATE2 structure to IDirectInputDevice::GetDeviceState().
	if( FAILED( hr = g_pJoystick->SetDataFormat( &c_dfDIJoystick2 ) ) )
		return hr;

	// Set the cooperative level to let DInput know how this device should
	// interact with the system and with other DInput applications.
	if( FAILED( hr = g_pJoystick->SetCooperativeLevel( hDlg, DISCL_EXCLUSIVE | 
		DISCL_FOREGROUND ) ) )
		return hr;

	// Enumerate the joystick objects. The callback function enabled user
	// interface elements for objects that are found, and sets the min/max
	// values property for discovered axes.
	if( FAILED( hr = g_pJoystick->EnumObjects( EnumObjectsCallback, 
		(VOID*)hDlg, DIDFT_ALL ) ) )
		return hr;

	return S_OK;
}

static BOOL CALLBACK EnumJoysticksCallback( const DIDEVICEINSTANCE* pdidInstance, VOID* pContext )
{
	DI_ENUM_CONTEXT* pEnumContext = (DI_ENUM_CONTEXT*) pContext;
	HRESULT hr;

	// Skip anything other than the perferred joystick device as defined by the control panel.  
	// Instead you could store all the enumerated joysticks and let the user pick.
	if( pEnumContext->bPreferredJoyCfgValid &&
		!IsEqualGUID( pdidInstance->guidInstance, pEnumContext->pPreferredJoyCfg->guidInstance ) )
		return DIENUM_CONTINUE;

	// Obtain an interface to the enumerated joystick.
	hr = g_pDI->CreateDevice( pdidInstance->guidInstance, &g_pJoystick, NULL );

	// If it failed, then we can't use this joystick. (Maybe the user unplugged
	// it while we were in the middle of enumerating it.)
	if( FAILED(hr) ) 
		return DIENUM_CONTINUE;

	// Stop enumeration. Note: we're just taking the first joystick we get. You
	// could store all the enumerated joysticks and let the user pick.
	return DIENUM_STOP;
}

static BOOL CALLBACK EnumObjectsCallback( const DIDEVICEOBJECTINSTANCE* pdidoi, VOID* pContext )
{
	HWND hDlg = (HWND)pContext;

	static int nSliderCount = 0;  // Number of returned slider controls
	static int nPOVCount = 0;     // Number of returned POV controls

	// For axes that are returned, set the DIPROP_RANGE property for the
	// enumerated axis in order to scale min/max values.
	if( pdidoi->dwType & DIDFT_AXIS )
	{
		DIPROPRANGE diprg; 
		diprg.diph.dwSize       = sizeof(DIPROPRANGE); 
		diprg.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
		diprg.diph.dwHow        = DIPH_BYID; 
		diprg.diph.dwObj        = pdidoi->dwType; // Specify the enumerated axis
		diprg.lMin              = -1000; 
		diprg.lMax              = +1000; 

		// Set the range for the axis
		if( FAILED( g_pJoystick->SetProperty( DIPROP_RANGE, &diprg.diph ) ) ) 
			return DIENUM_STOP;
	}

	return DIENUM_CONTINUE;
}

static HRESULT UpdateJoystick( HWND hDlg )
{
	HRESULT     hr;

	if( NULL == g_pJoystick ) 
		return S_OK;

	// Poll the device to read the current state
	hr = g_pJoystick->Poll(); 
	if( FAILED(hr) )  
	{
		// DInput is telling us that the input stream has been
		// interrupted. We aren't tracking any state between polls, so
		// we don't have any special reset that needs to be done. We
		// just re-acquire and try again.
		hr = g_pJoystick->Acquire();
		while( hr == DIERR_INPUTLOST ) 
			hr = g_pJoystick->Acquire();

		// hr may be DIERR_OTHERAPPHASPRIO or other errors.  This
		// may occur when the app is minimized or in the process of 
		// switching, so just try again later 
		return S_OK; 
	}

	// Get the input's device state
	if( FAILED( hr = g_pJoystick->GetDeviceState( sizeof(DIJOYSTATE2), &js ) ) )
		return hr; // The device should have been acquired during the Poll()

	return S_OK;
}

//////////////////////////////////////////////////////////////////////
// keyboard
static HRESULT CreateKeyboardDevice( HWND hDlg )
{
	HRESULT hr;
	DWORD   dwCoopFlags;

	if( bExclusive )
		dwCoopFlags = DISCL_EXCLUSIVE;
	else
		dwCoopFlags = DISCL_NONEXCLUSIVE;

	if( bForeground )
		dwCoopFlags |= DISCL_FOREGROUND;
	else
		dwCoopFlags |= DISCL_BACKGROUND;

	// Disabling the windows key is only allowed only if we are in foreground nonexclusive
	if( bDisableWindowsKey && !bExclusive && bForeground )
		dwCoopFlags |= DISCL_NOWINKEY;

	// Obtain an interface to the system keyboard device.
	if( FAILED( hr = g_pDI->CreateDevice( GUID_SysKeyboard, &g_pKeyboard, NULL ) ) )
		return hr;

	// Set the data format to "keyboard format" - a predefined data format 
	//
	// A data format specifies which controls on a device we
	// are interested in, and how they should be reported.
	//
	// This tells DirectInput that we will be passing an array
	// of 256 bytes to IDirectInputDevice::GetDeviceState.
	if( FAILED( hr = g_pKeyboard->SetDataFormat( &c_dfDIKeyboard ) ) )
		return hr;

	// Set the cooperativity level to let DirectInput know how
	// this device should interact with the system and with other
	// DirectInput applications.
	hr = g_pKeyboard->SetCooperativeLevel( hDlg, dwCoopFlags );
	if( hr == DIERR_UNSUPPORTED && !bForeground && bExclusive )
	{
		return S_OK;
	}

	if( FAILED(hr) )
		return hr;

	if( !bImmediate )
	{
		// IMPORTANT STEP TO USE BUFFERED DEVICE DATA!
		//
		// DirectInput uses unbuffered I/O (buffer size = 0) by default.
		// If you want to read buffered data, you need to set a nonzero
		// buffer size.
		//
		// Set the buffer size to DINPUT_BUFFERSIZE (defined above) elements.
		//
		// The buffer size is a DWORD property associated with the device.
		DIPROPDWORD dipdw;

		dipdw.diph.dwSize       = sizeof(DIPROPDWORD);
		dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
		dipdw.diph.dwObj        = 0;
		dipdw.diph.dwHow        = DIPH_DEVICE;
		dipdw.dwData            = SAMPLE_BUFFER_SIZE; // Arbitary buffer size

		if( FAILED( hr = g_pKeyboard->SetProperty( DIPROP_BUFFERSIZE, &dipdw.diph ) ) )
			return hr;
	}

	// Acquire the newly created device
	g_pKeyboard->Acquire();

	return S_OK;
}

static HRESULT ReadImmediateKeyboard( HWND hDlg )
{
	HRESULT hr;

	if( NULL == g_pKeyboard ) 
		return S_OK;

	// Get the input's device state, and put the state in dims
	ZeroMemory( diks, sizeof(diks) );
	hr = g_pKeyboard->GetDeviceState( sizeof(diks), diks );
	if( FAILED(hr) ) 
	{
		// DirectInput may be telling us that the input stream has been
		// interrupted.  We aren't tracking any state between polls, so
		// we don't have any special reset that needs to be done.
		// We just re-acquire and try again.

		// If input is lost then acquire and keep trying 
		hr = g_pKeyboard->Acquire();
		while( hr == DIERR_INPUTLOST ) 
			hr = g_pKeyboard->Acquire();

		// hr may be DIERR_OTHERAPPHASPRIO or other errors.  This
		// may occur when the app is minimized or in the process of 
		// switching, so just try again later 
		return S_OK; 
	}

	return S_OK;
}

static HRESULT ReadBufferedKeyboard( HWND hDlg )
{
	HRESULT            hr;

	if( NULL == g_pKeyboard ) 
		return S_OK;

	dwElementsKeybd = SAMPLE_BUFFER_SIZE;
	hr = g_pKeyboard->GetDeviceData( sizeof(DIDEVICEOBJECTDATA), didod_keybd, &dwElementsKeybd, 0 );
	if( hr != DI_OK ) 
	{
		// We got an error or we got DI_BUFFEROVERFLOW.
		//
		// Either way, it means that continuous contact with the
		// device has been lost, either due to an external
		// interruption, or because the buffer overflowed
		// and some events were lost.
		//
		// Consequently, if a button was pressed at the time
		// the buffer overflowed or the connection was broken,
		// the corresponding "up" message might have been lost.
		//
		// But since our simple sample doesn't actually have
		// any state associated with button up or down events,
		// there is no state to reset.  (In a real game, ignoring
		// the buffer overflow would result in the game thinking
		// a key was held down when in fact it isn't; it's just
		// that the "up" event got lost because the buffer
		// overflowed.)
		//
		// If we want to be cleverer, we could do a
		// GetDeviceState() and compare the current state
		// against the state we think the device is in,
		// and process all the states that are currently
		// different from our private state.
		hr = g_pKeyboard->Acquire();
		while( hr == DIERR_INPUTLOST ) 
			hr = g_pKeyboard->Acquire();

		// hr may be DIERR_OTHERAPPHASPRIO or other errors.  This
		// may occur when the app is minimized or in the process of 
		// switching, so just try again later 
		return S_OK; 
	}

	return S_OK;
}

//////////////////////////////////////////////////////////////////////////
// mouse
static HRESULT CreateMouseDevice( HWND hDlg )
{
	HRESULT hr;
	DWORD   dwCoopFlags;

	if( bExclusive )
		dwCoopFlags = DISCL_EXCLUSIVE;
	else
		dwCoopFlags = DISCL_NONEXCLUSIVE;

	if( bForeground )
		dwCoopFlags |= DISCL_FOREGROUND;
	else
		dwCoopFlags |= DISCL_BACKGROUND;

	// Create a DInput object
	if( FAILED( hr = DirectInput8Create( GetModuleHandle(NULL), DIRECTINPUT_VERSION, 
		IID_IDirectInput8, (VOID**)&g_pDI, NULL ) ) )
		return hr;

	// Obtain an interface to the system mouse device.
	if( FAILED( hr = g_pDI->CreateDevice( GUID_SysMouse, &g_pMouse, NULL ) ) )
		return hr;

	// Set the data format to "mouse format" - a predefined data format 
	//
	// A data format specifies which controls on a device we
	// are interested in, and how they should be reported.
	//
	// This tells DirectInput that we will be passing a
	// DIMOUSESTATE2 structure to IDirectInputDevice::GetDeviceState.
	if( FAILED( hr = g_pMouse->SetDataFormat( &c_dfDIMouse2 ) ) )
		return hr;

	// Set the cooperativity level to let DirectInput know how
	// this device should interact with the system and with other
	// DirectInput applications.
	hr = g_pMouse->SetCooperativeLevel( hDlg, dwCoopFlags );
	if( hr == DIERR_UNSUPPORTED && !bForeground && bExclusive )
	{
		return S_OK;
	}

	if( FAILED(hr) )
		return hr;

	if( !bImmediate )
	{
		// IMPORTANT STEP TO USE BUFFERED DEVICE DATA!
		//
		// DirectInput uses unbuffered I/O (buffer size = 0) by default.
		// If you want to read buffered data, you need to set a nonzero
		// buffer size.
		//
		// Set the buffer size to MOUSE_SAMPLE_BUFFER_SIZE (defined above) elements.
		//
		// The buffer size is a DWORD property associated with the device.
		DIPROPDWORD dipdw;
		dipdw.diph.dwSize       = sizeof(DIPROPDWORD);
		dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
		dipdw.diph.dwObj        = 0;
		dipdw.diph.dwHow        = DIPH_DEVICE;
		dipdw.dwData            = MOUSE_SAMPLE_BUFFER_SIZE; // Arbitary buffer size

		if( FAILED( hr = g_pMouse->SetProperty( DIPROP_BUFFERSIZE, &dipdw.diph ) ) )
			return hr;
	}

	// Acquire the newly created device
	g_pMouse->Acquire();

	return S_OK;
}

static HRESULT ReadImmediateMouse( HWND hDlg )
{
	HRESULT       hr;

	if( NULL == g_pMouse ) 
		return S_OK;

	// Get the input's device state, and put the state in dims
	ZeroMemory( &dims2, sizeof(dims2) );
	hr = g_pMouse->GetDeviceState( sizeof(DIMOUSESTATE2), &dims2 );
	if( FAILED(hr) ) 
	{
		// DirectInput may be telling us that the input stream has been
		// interrupted.  We aren't tracking any state between polls, so
		// we don't have any special reset that needs to be done.
		// We just re-acquire and try again.

		// If input is lost then acquire and keep trying 
		hr = g_pMouse->Acquire();
		while( hr == DIERR_INPUTLOST ) 
			hr = g_pMouse->Acquire();

		// hr may be DIERR_OTHERAPPHASPRIO or other errors.  This
		// may occur when the app is minimized or in the process of 
		// switching, so just try again later 
		return S_OK; 
	}

	return S_OK;
}

static HRESULT ReadBufferedMouse( HWND hDlg )
{
	// DIDEVICEOBJECTDATA didod[ MOUSE_SAMPLE_BUFFER_SIZE ];  // Receives buffered data 
	HRESULT            hr;

	if( NULL == g_pMouse ) 
		return S_OK;

	dwElementsMouse = MOUSE_SAMPLE_BUFFER_SIZE;
	hr = g_pMouse->GetDeviceData( sizeof(DIDEVICEOBJECTDATA), didod_mouse, &dwElementsMouse, 0 );
	if( hr != DI_OK ) 
	{
		// We got an error or we got DI_BUFFEROVERFLOW.
		//
		// Either way, it means that continuous contact with the
		// device has been lost, either due to an external
		// interruption, or because the buffer overflowed
		// and some events were lost.
		//
		// Consequently, if a button was pressed at the time
		// the buffer overflowed or the connection was broken,
		// the corresponding "up" message might have been lost.
		//
		// But since our simple sample doesn't actually have
		// any state associated with button up or down events,
		// there is no state to reset.  (In a real game, ignoring
		// the buffer overflow would result in the game thinking
		// a key was held down when in fact it isn't; it's just
		// that the "up" event got lost because the buffer
		// overflowed.)
		//
		// If we want to be cleverer, we could do a
		// GetDeviceState() and compare the current state
		// against the state we think the device is in,
		// and process all the states that are currently
		// different from our private state.
		hr = g_pMouse->Acquire();
		while( hr == DIERR_INPUTLOST ) 
			hr = g_pMouse->Acquire();

		// hr may be DIERR_OTHERAPPHASPRIO or other errors.  This
		// may occur when the app is minimized or in the process of 
		// switching, so just try again later 
		return S_OK; 
	}

	if( FAILED(hr) )  
		return hr;

	return S_OK;
}

//////////////////////////////////////////////////////////////////////////////
void UpdateInput(HWND hDlg)
{
	UpdateJoystick( hDlg );
	if (bImmediate)
	{
		ReadImmediateKeyboard( hDlg );
		ReadImmediateMouse( hDlg );
	}
	else
	{
		ReadBufferedKeyboard( hDlg );
		ReadBufferedMouse( hDlg );
	}
}

BYTE* getKeysStateImm()
{
	return diks;
}

DIMOUSESTATE2* getMouseStateImm()
{
	if( g_pMouse ) 
		return &dims2;
	
	return NULL;
}
	// if( g_pKeyboard ) 

DIJOYSTATE2* getJoystickState()
{
	if( g_pJoystick ) 
		return &js;
	
	return NULL;
}

DIDEVICEOBJECTDATA* getKeyStateBuff(int *ell)
{
	*ell = dwElementsKeybd;
	return didod_keybd;
}

DIDEVICEOBJECTDATA* getMouseStateBuff(int *ell)
{
	*ell = dwElementsMouse;
	return didod_mouse;
}

bool isKeyPressed(int key)
{
	return (diks[key] & 0x80) != 0x00;
}
