#ifndef __CLF_MATH__
#define __CLF_MATH__

#include <vector>
#include <list>

clFloat	const	pi	= 3.1415926535897932384626433832795f;
clFloat	const	tol = 0.0001f;		// clFloat type tolerance 

inline	clFloat	DegreesToRadians(clFloat deg) {
	return deg * pi / 180.0f;
}

inline	clFloat	RadiansToDegrees(clFloat rad) {	
	return rad * 180.0f / pi;
}

inline int dti(double d) {   // approximate clFloat to integer
	return (int)floor(d+0.5);
}

// angle in DEG !!!
clFloat myCos(clFloat a);
clFloat mySin(clFloat a);
clFloat myTan(clFloat a);
void init_math_lib();

//////////////////////////////////////////////////////////////////////////////////
/////////////// 2D section

class clf_point2D{
public:
    clFloat x;
    clFloat y;
    clf_point2D():x(0.f), y(0.f){
    }
    clf_point2D(clFloat _x, clFloat _y):x(_x), y(_y){
    }
    clf_point2D(int _x, int _y):x((clFloat)_x), y((clFloat)_y){
    }
    ~clf_point2D(){
    }
};

class clf_rect2D{
public:
    clFloat left;
    clFloat top;
    clFloat right;
    clFloat bottom;
    clf_rect2D() : left(0.f),top(0.f),right(0.f),bottom(0.f) {
    }
    clf_rect2D(clFloat l, clFloat t, clFloat r, clFloat b) : left(l),top(t),right(r),bottom(b) {
    }
    clf_rect2D(clf_rect2D& p) : left(p.left),top(p.top),right(p.right),bottom(p.bottom) {
    }
    ~clf_rect2D(){
    }
};

class clf_polygon2D : public std::vector< clf_point2D>{
public:
	clf_rect2D bounds;
	clf_polygon2D(){
	}
    ~clf_polygon2D(){
	}

	void add_point(const clf_point2D& pt){
		if (size() == 0)
		{
			bounds = clf_rect2D(pt.x, pt.y, pt.x, pt.y);
		}
		else
		{
			if (bounds.left > pt.x) bounds.left = pt.x;
			if (bounds.right < pt.x) bounds.right = pt.x;
			if (bounds.top < pt.y) bounds.top = pt.y;
			if (bounds.bottom > pt.y) bounds.bottom = pt.y;
		}
		push_back(clf_point2D(pt.x, pt.y));
	}
};

class clf_triangle2D
{
public:
	clf_triangle2D(){}
	clf_triangle2D(const clf_triangle2D& t){
		verts[0] = t.verts[0];
		verts[1] = t.verts[1];
		verts[2] = t.verts[2];
	}
	~clf_triangle2D(){}

	clf_point2D verts[3];
};

bool intersect(const clf_point2D& p1, const clf_point2D& p2, const clf_point2D& p3, const clf_point2D& p4, clf_point2D* p0);
bool point_in_polygon(clf_polygon2D& poly, clf_point2D pt);
bool polygons_collide(clf_polygon2D& poly1, clf_polygon2D& poly2);
void rotate2D(clFloat cx, clFloat cy, clFloat *x, clFloat *y, clFloat angle);
// check for intersection, but exclude the vertices
bool intersect_no_vertex(const clf_point2D& p1, const clf_point2D& p2, const clf_point2D& p3, const clf_point2D& p4, clf_point2D* p0);
// check if the line segment (p1,p2) is inside a polygon
bool line_in_polygon(clf_polygon2D& poly, clf_point2D& p1, clf_point2D& p2);
bool triangulate(clf_polygon2D& poly, std::vector<clf_triangle2D>& result);

//////////////////////////////////////////////////////////////////////////////////
/////////////// 3D section

class clf_matrix4x4;

class clf_vector3D{
public:
	clFloat x,y,z;
	clf_vector3D(){
		x = y = z = 0;
	}
	clf_vector3D(clFloat _x, clFloat _y, clFloat _z):x(_x), y(_y), z(_z){
	}
	clf_vector3D(const clf_vector3D& v):x(v.x), y(v.y), z(v.z){
	}
	clf_vector3D& operator=(const clf_vector3D& v){
		x = v.x; y = v.y; z = v.z;
		return *this;
	}
	clf_vector3D& operator*=(clFloat s){	// scalar multiply
		x *= s; y *= s; z *= s;
		return *this;
	}
	clf_vector3D& operator/=(clFloat s){	// scalar divide
		x /= s; y /= s; z /= s;
		return *this;
	}
	clf_vector3D operator-(void) {
		return clf_vector3D(-x, -y, -z);
	}
	operator clFloat* () {
		return &x;
	}
	clf_vector3D& scale(clFloat s){
		x *= s; y *= s; z *= s;
		return *this;
	}
	clFloat length(){
		return sqrt(x*x + y*y + z*z);
	}
	void normalize(){ 
		clFloat fLength = length();
		if (fLength > 0.0001)
			fLength = 1.0f / fLength;
		scale(fLength);
	}
	clf_vector3D inverse(){
		clf_vector3D r(-x, -y, -z);
		return r;
	}
	clf_vector3D& rotate(clf_matrix4x4& mat);

};

inline clf_vector3D operator + (clf_vector3D& m1, clf_vector3D& m2){
	clf_vector3D r(m1);
	r.x += m2.x; r.y += m2.y; r.z += m2.z;
	return r;
}
/* */
inline clf_vector3D operator - (const clf_vector3D& m1, const clf_vector3D& m2){
	clf_vector3D r(m1);
	r.x -= m2.x; r.y -= m2.y; r.z -= m2.z;
	return r;
}
/* */
inline clFloat dotProduct(const clf_vector3D& v1, const clf_vector3D& v2){
	return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z; 
}
inline clf_vector3D crossProduct(const clf_vector3D& v1, const clf_vector3D& v2){
	clf_vector3D vResult;
	vResult.x = v1.y*v2.z - v2.y*v1.z;
	vResult.y = -v1.x*v2.z + v2.x*v1.z;
	vResult.z = v1.x*v2.y - v2.x*v1.y;
	return vResult;
}
inline clf_vector3D operator*(const clf_vector3D& v1, const clf_vector3D& v2){
	clf_vector3D vResult;
	vResult.x = v1.y*v2.z - v2.y*v1.z;
	vResult.y = -v1.x*v2.z + v2.x*v1.z;
	vResult.z = v1.x*v2.y - v2.x*v1.y;
	return vResult;
}
inline clf_vector3D operator*(clf_vector3D& v1, clFloat s){
	// scalar multiply
	clf_vector3D r(v1);
	r.x *= s; r.y *= s; r.z *= s;
	return r;
}													 
inline clf_vector3D operator*(clFloat s, clf_vector3D& v1)
{	// scalar multiply
	clf_vector3D r(v1);
	r.x *= s; r.y *= s; r.z *= s;
	return r;
}													 
inline clf_vector3D operator/(clf_vector3D& v1, clFloat s)
{	// scalar multiply
	clf_vector3D r(v1);
	r.x /= s; r.y /= s; r.z /= s;
	return r;
}													 

class clf_vector4 {
	clFloat data[4];
public:
	clf_vector4(){}
	clf_vector4(const clf_vector4& m1){
		memcpy(data, m1.data, 4*sizeof(clFloat));
	}
	clf_vector4(clFloat x, clFloat y, clFloat z,clFloat a){
		data[0] = x;
		data[1] = y;
		data[2] = z;
		data[3] = a;
	}
	virtual ~clf_vector4(){}
	virtual clFloat& operator[](int index){
		return data[index];
	}
	virtual clFloat operator[](int index) const{
		return data[index];
	}
	clf_vector4& operator = (clf_vector4& m1){
		memcpy(data, m1.data, 4*sizeof(clFloat));
		return *this;
	}
	operator clFloat* () {
		return data;
	}
	clf_vector4& scale(clFloat sc){
		for (int i=0; i<4; ++i)
			data[i] *= sc;
		return *this;
	}
	clf_vector4 operator + (clf_vector4& m1){
		clf_vector4 r;
		for (int i=0; i<4; ++i)
			r[i] = data[i] + m1.data[i];
		return r;
	}
	clf_vector4 operator - (clf_vector4& m1){
		clf_vector4 r;
		for (int i=0; i<4; ++i)
			r[i] = data[i] - m1.data[i];
		return r;
	}
};

class Matrix3x3 {
public:
	// elements eij: i -> row, j -> column
	clFloat	e11, e12, e13, e21, e22, e23, e31, e32, e33;	
	Matrix3x3(void){
		e11 = e12 = e13 = 0;
		e21 = e22 = e23 = 0;
		e31 = e32 = e33 = 0;
	}
	Matrix3x3(	clFloat r1c1, clFloat r1c2, clFloat r1c3, clFloat r2c1, clFloat r2c2, clFloat r2c3, clFloat r3c1, clFloat r3c2, clFloat r3c3 ){
		e11 = r1c1;e12 = r1c2;e13 = r1c3;
		e21 = r2c1;e22 = r2c2;e23 = r2c3;
		e31 = r3c1;e32 = r3c2;e33 = r3c3;
	}
	clFloat	det(void){
		return	e11*e22*e33 - e11*e32*e23 + e21*e32*e13 - e21*e12*e33 + e31*e12*e23 - e31*e22*e13;	
	}
	Matrix3x3	Transpose(void){
		return Matrix3x3(e11,e21,e31,e12,e22,e32,e13,e23,e33);
	}
	Matrix3x3	Inverse(void){
		clFloat	d = e11*e22*e33 - e11*e32*e23 + e21*e32*e13 - e21*e12*e33 + e31*e12*e23 - e31*e22*e13;
		if (d == 0) d = 1;
		return	Matrix3x3(	(e22*e33-e23*e32)/d,-(e12*e33-e13*e32)/d,(e12*e23-e13*e22)/d,
			-(e21*e33-e23*e31)/d,(e11*e33-e13*e31)/d,-(e11*e23-e13*e21)/d,
			(e21*e32-e22*e31)/d,-(e11*e32-e12*e31)/d,(e11*e22-e12*e21)/d );	
	}
	Matrix3x3& operator+=(Matrix3x3 m){
		e11 += m.e11;e12 += m.e12;e13 += m.e13;
		e21 += m.e21;e22 += m.e22;e23 += m.e23;
		e31 += m.e31;e32 += m.e32;e33 += m.e33;
		return *this;
	}
	Matrix3x3& operator-=(Matrix3x3 m){
		e11 -= m.e11;e12 -= m.e12;e13 -= m.e13;
		e21 -= m.e21;e22 -= m.e22;e23 -= m.e23;
		e31 -= m.e31;e32 -= m.e32;e33 -= m.e33;
		return *this;
	}
	Matrix3x3& operator*=(clFloat s){
		e11 *= s;e12 *= s;e13 *= s;
		e21 *= s;e22 *= s;e23 *= s;
		e31 *= s;e32 *= s;e33 *= s;
		return *this;
	}
	Matrix3x3& operator/=(clFloat s){
		e11 /= s;e12 /= s;e13 /= s;
		e21 /= s;e22 /= s;e23 /= s;
		e31 /= s;e32 /= s;e33 /= s;
		return *this;
	}
};

inline	Matrix3x3 operator+(Matrix3x3 m1, Matrix3x3 m2)
{
	return	Matrix3x3(	m1.e11+m2.e11,m1.e12+m2.e12,m1.e13+m2.e13,
						m1.e21+m2.e21,m1.e22+m2.e22,m1.e23+m2.e23,
						m1.e31+m2.e31,m1.e32+m2.e32,m1.e33+m2.e33);
}

inline	Matrix3x3 operator-(Matrix3x3 m1, Matrix3x3 m2)
{
	return	Matrix3x3(	m1.e11-m2.e11,m1.e12-m2.e12,m1.e13-m2.e13,
						m1.e21-m2.e21,m1.e22-m2.e22,m1.e23-m2.e23,
						m1.e31-m2.e31,m1.e32-m2.e32,m1.e33-m2.e33);
}

inline	Matrix3x3 operator/(Matrix3x3 m, clFloat s)
{	
	return	Matrix3x3(	m.e11/s,m.e12/s,m.e13/s,
						m.e21/s,m.e22/s,m.e23/s,
						m.e31/s,m.e32/s,m.e33/s);
}

inline	Matrix3x3 operator*(Matrix3x3 m1, Matrix3x3 m2)
{
	return Matrix3x3(	m1.e11*m2.e11 + m1.e12*m2.e21 + m1.e13*m2.e31,
						m1.e11*m2.e12 + m1.e12*m2.e22 + m1.e13*m2.e32,
						m1.e11*m2.e13 + m1.e12*m2.e23 + m1.e13*m2.e33,
						m1.e21*m2.e11 + m1.e22*m2.e21 + m1.e23*m2.e31,
						m1.e21*m2.e12 + m1.e22*m2.e22 + m1.e23*m2.e32,
						m1.e21*m2.e13 + m1.e22*m2.e23 + m1.e23*m2.e33,
						m1.e31*m2.e11 + m1.e32*m2.e21 + m1.e33*m2.e31,
						m1.e31*m2.e12 + m1.e32*m2.e22 + m1.e33*m2.e32,
						m1.e31*m2.e13 + m1.e32*m2.e23 + m1.e33*m2.e33 );
}

inline	Matrix3x3 operator*(Matrix3x3 m, clFloat s)
{
	return	Matrix3x3(	m.e11*s,m.e12*s,m.e13*s,
						m.e21*s,m.e22*s,m.e23*s,
						m.e31*s,m.e32*s,m.e33*s);
}

inline	Matrix3x3 operator*(clFloat s, Matrix3x3 m)
{
	return	Matrix3x3(	m.e11*s,m.e12*s,m.e13*s,
						m.e21*s,m.e22*s,m.e23*s,
						m.e31*s,m.e32*s,m.e33*s);
}

inline	clf_vector3D operator*(Matrix3x3 m, clf_vector3D u)
{
	return clf_vector3D(	m.e11*u.x + m.e12*u.y + m.e13*u.z,
					m.e21*u.x + m.e22*u.y + m.e23*u.z,
					m.e31*u.x + m.e32*u.y + m.e33*u.z);					
}

inline	clf_vector3D operator*(clf_vector3D u, Matrix3x3 m)
{
	return clf_vector3D(	u.x*m.e11 + u.y*m.e21 + u.z*m.e31,
					u.x*m.e12 + u.y*m.e22 + u.z*m.e32,
					u.x*m.e13 + u.y*m.e23 + u.z*m.e33);
}


class clf_matrix4x4 {
	clFloat data[16];
	clFloat DetIJ(int i, int j);
public:
	clf_matrix4x4(){
		memset(data,0,sizeof(clFloat)*16);
	}
	clf_matrix4x4(const clf_matrix4x4& m1){
		memcpy(data, m1.data, 16*sizeof(clFloat));
	}
	virtual ~clf_matrix4x4(){}
	virtual clFloat& operator[](int index){
		return data[index];
	}
	virtual clFloat operator[](int index) const{
		return data[index];
	}
	clf_matrix4x4& operator = (const clf_matrix4x4& m1){
		memcpy(data, m1.data, 16*sizeof(clFloat));
		return *this;
	}
	operator clFloat* () {
		return data;
	}
	clf_matrix4x4& scale(clFloat sc){
		for (int i=0; i<16; ++i)
			data[i] *= sc;
		return *this;
	}
	clf_matrix4x4 operator + (clf_matrix4x4& m1){
		clf_matrix4x4 r;
		for (int i=0; i<16; ++i)
			r[i] = data[i] + m1.data[i];
		return r;
	}
	clf_matrix4x4 operator - (clf_matrix4x4& m1){
		clf_matrix4x4 r;
		for (int i=0; i<16; ++i)
			r[i] = data[i] - m1.data[i];
		return r;
	}
	clf_matrix4x4 operator *(clFloat sc){
		clf_matrix4x4 r;
		for (int i=0; i<16; ++i)
			r[i] = data[i] * sc;
		return r;
	}
	clf_vector4 operator *(clf_vector4& v);

	void loadIdentity();
	clf_matrix4x4 operator *(const clf_matrix4x4& m2);
	void Transpose();
	clf_matrix4x4& Invert();
	void TranslationMatrix(clFloat x, clFloat y, clFloat z);
	void ScalingMatrix(clFloat x, clFloat y, clFloat z);
	void RotationMatrix(clFloat angle, clFloat x, clFloat y, clFloat z);
	void RotationMatrix(clFloat angle, clf_vector3D& vAxis);
	void MakeShadowMatrix(clf_vector3D vPoints[3], clf_vector4& vLightPos);
};

void clf_getNormalVector(clf_vector3D& vP1, clf_vector3D& vP2, clf_vector3D& vP3, clf_vector3D& vNormal);
void clf_rotateVector(const clf_vector3D& vSrcVector, const clf_matrix4x4& mMatrix, clf_vector3D& vOut);
void clf_unrotateVector(const clf_vector3D& vSrcVector, const clf_matrix4x4& mMatrix, clf_vector3D& vOut);
/*
dVector dMatrix::UnrotateVector (const dVector &v) const
{
	return dVector (v % m_front, v % m_up, v % m_right);
}
*/

void clf_transformPoint(const clf_vector3D& vSrcVector, const clf_matrix4x4& mMatrix, clf_vector3D& vOut);
void clf_getPlaneEquation(clf_vector3D& vPoint1, clf_vector3D& vPoint2, clf_vector3D& vPoint3, clf_vector4& vPlane);
clFloat clf_distanceToPlane(clf_vector3D& vPoint, clf_vector4& vPlane);

void rotate_point(clFloat x, clFloat y, clFloat z, clf_vector3D& point);
void rotate_point_rad(clFloat x, clFloat y, clFloat z, clf_vector3D& point);
void rotate_f(clFloat x, clFloat y, clFloat z, clf_vector3D& point);


#endif // __CLF_MATH__
