#ifndef __CLF_WORLD__
#define __CLF_WORLD__

#include "clf_math.h"

// base object type
class clf_world_object
{
public:
	static clf_vector3D gravity; // the famous g

	clFloat m_mass;				// object mass
	clFloat m_radius;     // bounding sphere
	clFloat m_radius2;		// radius squared to speed up distance calculations
	clf_vector3D m_pos; // position
	clf_vector3D m_vel; // velocity
	bool m_fixed_pos;

	clf_world_object(const clFloat m, const clFloat r );
	virtual ~clf_world_object();
	virtual void step_simulation(clFloat fElapsed);
	virtual void render();
	// check for collision pure virtual function
	virtual bool collides_with(clf_world_object* pob)=0;
	virtual bool is_at_xz(clf_vector3D& pos, clFloat r);
};

// our game world
class clf_world
{
	static clFloat sim_speed; // used to control game speed, 1 is realtime <1 slow motion, >1 fast
public:
	// a list of the world objects
	std::list<clf_world_object*> m_objects;
	clf_world();
	~clf_world();
	// iterate the objects and find collisions
	void collision_detection();
	// calculate next frame
	void step_simulation(clFloat fElapsed);
	// draw the objects
	void render();
	// add object to world
	void add_object(clf_world_object* pob);

	clf_world_object* object_at(clf_vector3D& pos, clFloat r);
};

// the sphere object
class clf_sphere : public clf_world_object
{
public:
	clf_sphere(clFloat m, clFloat r );
	virtual ~clf_sphere();
	virtual void render();
	virtual void step_simulation(clFloat fElapsed);
	void collision_respond(clf_sphere* pob);
	virtual bool collides_with(clf_sphere* pob);
	virtual bool collides_with(clf_world_object* pob);
};

class clf_ground : public clf_world_object
{
	clFloat north, south, east, west;
public:
	clf_ground():clf_world_object(1.e30f, 0){
		m_fixed_pos = true;
		north = 100.f;
		south = -100.f;
		east = 100.f;
		west = -100.f;
	}
	virtual ~clf_ground(){
	}
	virtual void render(){
		// glColor4f(.3f, .8f, .0f, 1.f);
		glColor4f(.1f, .1f, .1f, 1.f);
		glBegin(GL_QUADS);
		glNormal3f(0,1,0);
		glVertex3f(south, 0, west);
		glVertex3f(south, 0, east);
		glVertex3f(north, 0, east);
		glVertex3f(north, 0, west);
		glEnd();
	}
	virtual bool collides_with(clf_world_object* pob){
		return false;
	}
	void set_extend(clFloat n, clFloat s, clFloat e, clFloat w){
		north = n;
		south = s;
		east = e;
		west = w;
	}
};

class building : public clf_world_object
{
public:
	int m_tex;
	clf_vector3D m_size;  // size
	building(const clf_vector3D& p, const clf_vector3D& s):clf_world_object(1.e30f, 0) {
		m_tex = -1;
		m_pos = p; // position
		m_size = s; 
		m_fixed_pos = true;
	}
	virtual ~building(){
	}
	void set_texture(int tex){
		m_tex = tex;
	}
	virtual void render();
	virtual bool collides_with(clf_world_object* pob){
		return false;
	}
	virtual bool is_at_xz(clf_vector3D& pos, clFloat r){
		clFloat xmin = m_pos.x-r-m_size.x/2.f;
		clFloat xmax = m_pos.x+r+m_size.x/2.f;
		clFloat zmin = m_pos.z-r-m_size.z/2.f;
		clFloat zmax = m_pos.z+r+m_size.z/2.f;
		if (xmin > pos.x)
			return false;
		if (xmax < pos.x)
			return false;
		if (zmin > pos.z)
			return false;
		if (zmax < pos.z)
			return false;
		return true;
	}
};

#endif // __CLF_WORLD__